@@ -309,16 +309,16 @@ class CursorHandler(ClangHandler):
             # but we need to put a CFUNCTYPE as a value of the name variable
             init_value = _type
         elif self.is_array_type(_ctype):
-            # an integer litteral will be the size
-            # an string litteral will be the value
+            # an integer literal will be the size
+            # a string literal will be the value
             # any list member will be children of a init_list_expr
             # FIXME Move that code into typedesc
             def countof(k, l):
                 return [item[0] for item in l].count(k)
 
-            if (countof(CursorKind.INIT_LIST_EXPR, init_value) == 1):
+            if countof(CursorKind.INIT_LIST_EXPR, init_value) == 1:
                 init_value = dict(init_value)[CursorKind.INIT_LIST_EXPR]
-            elif (countof(CursorKind.STRING_LITERAL, init_value) == 1):
+            elif countof(CursorKind.STRING_LITERAL, init_value) == 1:
                 # we have a initialised c_array
                 init_value = dict(init_value)[CursorKind.STRING_LITERAL]
             else:
@@ -383,8 +383,7 @@ class CursorHandler(ClangHandler):
         # but really it depends...
         if child.kind.is_unexposed():
             # recurse until we find a literal kind
-            init_value = self._get_var_decl_init_value(_ctype,
-                                                       child.get_children())
+            init_value = self._get_var_decl_init_value(_ctype, child.get_children())
             if len(init_value) == 0:
                 init_value = None
             elif len(init_value) == 1:
@@ -392,18 +391,16 @@ class CursorHandler(ClangHandler):
             else:
                 log.error('_get_var_decl_init_value_single: Unhandled case')
                 assert len(init_value) <= 1
+        # elif child.kind == CursorKind.STRING_LITERAL:
+        #     _v = self._literal_handling(child)
+        #     init_value = (child.kind, _v)
         else:  # literal or others
             _v = self.parse_cursor(child)
-            if isinstance(
-                    _v, list) and child.kind not in [CursorKind.INIT_LIST_EXPR, CursorKind.STRING_LITERAL]:
-                log.warning(
-                    '_get_var_decl_init_value_single: TOKENIZATION BUG CHECK: %s',
-                    _v)
+            if isinstance(_v, list) and len(_v) > 0 and child.kind not in [CursorKind.INIT_LIST_EXPR, CursorKind.STRING_LITERAL]:
+                log.warning('_get_var_decl_init_value_single: TOKENIZATION BUG CHECK: %s', _v)
                 _v = _v[0]
             init_value = (child.kind, _v)
-        log.debug(
-            '_get_var_decl_init_value_single: returns %s',
-            str(init_value))
+        log.debug('_get_var_decl_init_value_single: returns %s', str(init_value))
         return init_value
 
     def _clean_string_literal(self, cursor, value):
@@ -457,14 +454,18 @@ class CursorHandler(ClangHandler):
             # init_value = ' '.join([t.spelling for t in children[0].get_tokens()
             # if t.spelling != ';'])
         because some literal might need cleaning."""
-        # use a shortcut - does not work on unicode var_decl
-        # if cursor.kind == CursorKind.STRING_LITERAL:
-        #     value = cursor.displayname
-        #     value = self._clean_string_literal(cursor, value)
-        #     return value
         tokens = list(cursor.get_tokens())
-        log.debug('literal has %d tokens.[ %s ]', len(tokens),
-                  str([str(t.spelling) for t in tokens]))
+        log.debug('literal has %d tokens.[ %s ]', len(tokens), ' '.join([str(t.spelling) for t in tokens]))
+        if len(tokens) == 1 and cursor.kind == CursorKind.STRING_LITERAL:
+            # use a shortcut that works for unicode
+            value = tokens[0].spelling
+            value = self._clean_string_literal(cursor, value)
+            return value
+        elif cursor.kind == CursorKind.STRING_LITERAL:
+            # use a shortcut - does not work on unicode var_decl
+            value = cursor.displayname
+            value = self._clean_string_literal(cursor, value)
+            return value
         final_value = []
         # code.interact(local=locals())
         log.debug('cursor.type:%s', cursor.type.kind.name)
@@ -485,10 +486,8 @@ class CursorHandler(ClangHandler):
                 continue
             # elif token.cursor.kind == CursorKind.VAR_DECL:
             elif token.location not in cursor.extent:
-                log.debug(
-                    'FIXME BUG: token.location not in cursor.extent %s',
-                    value)
-                # FIXME
+                # log.debug('FIXME BUG: token.location not in cursor.extent %s', value)
+                # 2021 clang 11, this seems fixed ?
                 # there is most probably a BUG in clang or python-clang
                 # when on #define with no value, a token is taken from
                 # next line. Which break stuff.
@@ -521,6 +520,7 @@ class CursorHandler(ClangHandler):
                 value = self.get_registered(value).body
                 # already cleaned value = self._clean_string_literal(token.cursor, value)
             elif token.cursor.kind == CursorKind.MACRO_DEFINITION:
+                tk = token.kind
                 if i == 0:
                     # ignore, macro name
                     pass
@@ -528,18 +528,33 @@ class CursorHandler(ClangHandler):
                     # and just clean it
                     value = self._clean_string_literal(token.cursor, value)
                 elif token.kind == TokenKind.IDENTIFIER:
-                    # parse that, try to see if there is another Macro in there.
-                    value = self.get_registered(value).body
+                    # log.debug("Ignored MACRO_DEFINITION token identifier : %s", value)
+                    # Identifier in Macro... Not sure what to do with that.
+                    if self.is_registered(value):
+                        # parse that, try to see if there is another Macro in there.
+                        value = self.get_registered(value).body
+                        log.debug("Found MACRO_DEFINITION token identifier : %s", value)
+                    else:
+                        value = typedesc.UndefinedIdentifier(value)
+                        log.debug("Undefined MACRO_DEFINITION token identifier : %s", value)
+                    pass
+                elif token.kind in [TokenKind.COMMENT, TokenKind.KEYWORD, TokenKind.PUNCTUATION]:
+                    # log.debug("Ignored MACRO_DEFINITION token.kind: %s", token.kind.name)
+                    pass
 
             # add token
-            final_value.append(value)
+            if value is not None:
+                final_value.append(value)
         # return the EXPR
         # code.interact(local=locals())
+        # FIXME, that will break. We need constant type return
         if len(final_value) == 1:
             return final_value[0]
         # Macro definition of a string using multiple macro
-        if isinstance(final_value, list) and cursor.kind == CursorKind.STRING_LITERAL:
-            final_value = ''.join(final_value)
+        if isinstance(final_value, list):
+            if cursor.kind == CursorKind.STRING_LITERAL:
+                final_value = ''.join(final_value)
+        log.debug('_literal_handling final_value: %s', final_value)
         return final_value
 
     INTEGER_LITERAL = _literal_handling
@@ -553,6 +568,24 @@ class CursorHandler(ClangHandler):
         """Returns a string with the literal that are part of the operation."""
         values = self._literal_handling(cursor)
         retval = ''.join([str(val) for val in values])
+        log.debug('cursor.type.kind:%s', cursor.type.kind.name)
+        if cursor.kind == CursorKind.UNARY_OPERATOR:
+            if cursor.type.kind in [TypeKind.INT, TypeKind.LONG]:
+                if '0x' in retval:
+                    retval = int(retval, 16)
+                else:
+                    try:
+                        retval = int(retval)
+                    except ValueError:
+                        # fall back on pass through
+                        pass
+            elif cursor.type.kind in [TypeKind.FLOAT, TypeKind.DOUBLE]:
+                retval = float(retval)
+        # Things we do not want to do:
+        # elif cursor.kind == CursorKind.BINARY_OPERATOR:
+        #     # cursor.kind == binary_operator, then need to make some additions
+        #     retval = eval(retval)
+
         return retval
 
     UNARY_OPERATOR = _operator_handling
@@ -1033,28 +1066,48 @@ class CursorHandler(ClangHandler):
         """
         Parse MACRO_DEFINITION, only present if the TranslationUnit is
         used with TranslationUnit.PARSE_DETAILED_PROCESSING_RECORD.
+        By default, macro are not parsed. requires -k m || parser.activate_macros_parsing()
         """
-        # TODO: optionalize macro parsing. It takes a LOT of time.
+        # macro parsing takes a LOT of time.
         # ignore system macro
         if (not hasattr(cursor, 'location') or cursor.location is None or
                 cursor.location.file is None):
+            # keep track of sizes from clang directly
+            # but we already did that elsewhere in clangparser.py make_ctypes_convertor
+            # if cursor.displayname.startswith('__SIZEOF_'):
+            #     typ = cursor.displayname[len('__SIZEOF_'):-2]
+            #     self.__sizeof[typ] = list(cursor.get_tokens())[1].spelling
             return False
         name = self.get_unique_name(cursor)
-        # if name == 'A':
-        #    code.interact(local=locals())
-        # Tokens !!! .kind = {IDENTIFIER, KEYWORD, LITERAL, PUNCTUATION,
-        # COMMENT ? } etc. see TokenKinds.def
+        # MACRO_DEFINITION are a list of Tokens
+        # .kind = {IDENTIFIER, KEYWORD, LITERAL, PUNCTUATION, COMMENT ? }
         comment = None
         tokens = self._literal_handling(cursor)
         # Macro name is tokens[0]
         # get Macro value(s)
         value = True
+        # args should be filled when () are in tokens,
+        args = None
         if isinstance(tokens, list):
             if len(tokens) == 2:
+                # #define key value
                 value = tokens[1]
+            elif tokens[1] == '(':
+                # function macro
+                str_tokens = [str(_) for _ in tokens[1:tokens.index(')')+1]]
+                args = ''.join(str_tokens).replace(',', ', ')
+                str_tokens = [str(_) for _ in tokens[tokens.index(')')+1:]]
+                value = ''.join(str_tokens)
+            elif len(tokens) > 2:
+                # #define key a b c
+                value = list(tokens[1:])
             else:
+                # FIXME no reach ?!
                 # just merge the list of tokens
-                value = ''.join(tokens[1:])
+                value = ' '.join(tokens[1:])
+        elif isinstance(tokens, str):
+            # #define only
+            value = True
         # macro comment maybe in tokens. Not in cursor.raw_comment
         for t in cursor.get_tokens():
             if t.kind == TokenKind.COMMENT:
@@ -1064,8 +1117,8 @@ class CursorHandler(ClangHandler):
         # why not Cursor.kind GNU_NULL_EXPR child instead of a token ?
         if name == 'NULL' or value == '__null':
             value = None
-        log.debug('MACRO: #define %s %s', tokens[0], value)
-        obj = typedesc.Macro(name, None, value)
+        log.debug('MACRO: #define %s%s %s', name, args or '', value)
+        obj = typedesc.Macro(name, args, value)
         try:
             self.register(name, obj)
         except DuplicateDefinitionException:
@@ -1079,10 +1132,22 @@ class CursorHandler(ClangHandler):
         obj.comment = comment
         return True
 
-    # @log_entity
-    # def MACRO_INSTANTIATION(self, cursor):
-    #     log.debug("I was here")
-    #     self.set_location(obj, cursor)
-    #     # set the comment in the obj
-    #     obj.comment = comment
-    #     return True
+    @log_entity
+    def MACRO_INSTANTIATION(self, cursor):
+        """We could use this to count instantiations
+        so we now, if we need to generate python code or comment for this macro ? """
+        log.debug('cursor.spelling: %s', cursor.spelling)
+        log.debug('cursor.kind: %s', cursor.kind.name)
+        log.debug('cursor.type.kind: %s', cursor.type.kind.name)
+        # no children ?
+        for child in cursor.get_children():
+            log.debug('child.spelling: %s', child.spelling)
+            log.debug('child.kind: %s', child.kind.name)
+            log.debug('child.type.kind: %s', child.type.kind.name)
+
+            # ret.append(self.parse_cursor(child))
+        # log.debug('cursor.type:%s', cursor.type.kind.name)
+        # self.set_location(obj, cursor)
+        # set the comment in the obj
+        # obj.comment = comment
+        return True
